"""MQPMO: Put Message Options"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, ibmmqc

class PMO(MQOpts):
    """ Construct an MQPMO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [
            ['_StrucId', CMQC.MQPMO_STRUC_ID, '4s'],
            ['Version', CMQC.MQPMO_VERSION_3, MQLONG_TYPE],
            ['Options', CMQC.MQPMO_NONE, MQLONG_TYPE],
            ['Timeout', -1, MQLONG_TYPE],
            ['Context', 0, MQLONG_TYPE],
            ['KnownDestCount', 0, MQLONG_TYPE],
            ['UnknownDestCount', 0, MQLONG_TYPE],
            ['InvalidDestCount', 0, MQLONG_TYPE],
            ['ResolvedQName', b'', '48s'],
            ['ResolvedQMgrName', b'', '48s'],
            ['RecsPresent', 0, MQLONG_TYPE],
            ['PutMsgRecFields', 0, MQLONG_TYPE],
            ['PutMsgRecOffset', 0, MQLONG_TYPE],
            ['ResponseRecOffset', 0, MQLONG_TYPE],
            ['PutMsgRecPtr', 0, 'P'],
            ['ResponseRecPtr', 0, 'P']]

        pmo_current_version = ibmmqc.__strucversions__.get("pmo", 1)
        if pmo_current_version >= CMQC.MQPMO_VERSION_3:
            opts += [
                ['OriginalMsgHandle', 0, 'q'],
                ['NewMsgHandle', 0, 'q'],
                ['Action', 0, MQLONG_TYPE],
                ['PubLevel', 9, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)


# Backward compatibility
pmo = PMO  # pylint: disable=invalid-name




"""Setup script usable for setuptools"""
#
# Original Author: Maas-Maarten Zeeman
# Rewritten to remove deprecated distutils: Mark Taylor
#

import os
import platform

from shutil import which
from struct import calcsize

from setuptools import setup, Extension

# This is the only place where package version number is set!
# The version should correspond to PEP440 and gets normalised if
# not in the right format. VRM can be followed with a|b|rc with a further numeric
# to indicate alpha/beta/release-candidate versions.
VERSION = '2.0.2'

_ABI_LIMITS = {
    # Minimum Python version that this package supports.
    'python_requires': ">=3.9",
    # Used as a macro.
    'Py_LIMITED_API': 0x03090000,
    # Used for bdist_wheel option.
    "py_limited_api": "cp39",
}

# If the MQ SDK is in a non-default location, set MQ_FILE_PATH environment variable.
custom_path = os.environ.get('MQ_FILE_PATH', None)

# Always build in 64-bit mode. And use libmqm regardless of whether the package
# will be used in client or local bindings mode. Since V7 the one library can handle
# both modes.
if calcsize('P') != 8:
    # Seems the closest exception (now an alias of EnvironmentError)
    raise OSError("Cannot build in 32-bit mode. Only 64-bit systems supported.")

# The include_dirs and library_dirs used to be lists that always included the default
# directories. But if you've gone to the trouble of setting a non-default
# custom_path, then we should only use that. Otherwise the compile might get confused and use
# the "wrong" version of the files.

def get_windows_settings():
    """ Windows settings.
    """

    include_dirs = [r'c:\Program Files\IBM\MQ\tools\c\include']
    library_dirs = [r'c:\Program Files\IBM\MQ\tools\Lib64']
    if custom_path:
        library_dirs = [r'{}\tools\Lib64'.format(custom_path)]
        include_dirs = [r'{}\tools\c\include'.format(custom_path)]

    libraries = ['mqm']

    return library_dirs, include_dirs, libraries

def get_aix_settings():
    """ AIX settings.
    """

    library_dirs = ['/usr/mqm/lib64']
    include_dirs = ['/usr/mqm/inc']

    if custom_path:
        library_dirs = ['{}/lib64'.format(custom_path)]
        include_dirs = ['{}/inc'.format(custom_path)]

    libraries = ['mqm_r']

    return library_dirs, include_dirs, libraries

def get_generic_unix_settings():
    """ Generic UNIX, including Linux, settings.
    """

    library_dirs = ['/opt/mqm/lib64']
    include_dirs = ['/opt/mqm/inc']

    if custom_path:
        library_dirs = ['{}/lib64'.format(custom_path)]
        include_dirs = ['{}/inc'.format(custom_path)]

    # Get an embedded rpath into the library which can reduce need for LD_LIBRARY_PATH
    for d in library_dirs:
        ld_flags.append("-Wl,-rpath," + d)
    libraries = ['mqm_r']

    return library_dirs, include_dirs, libraries

def get_locations_by_command_path(command_path):
    """ Extracts directory locations by the path to one of MQ commands, such as dspmqver.
    """
    command_dir = os.path.dirname(command_path)
    mq_installation_path = os.path.abspath(os.path.join(command_dir, '..'))

    library_dirs = ['{}/lib64'.format(mq_installation_path)]
    include_dirs = ['{}/inc'.format(mq_installation_path)]

    libraries = ['mqm_r']

    return library_dirs, include_dirs, libraries


# Define the C part(s) here. So we can potentially modify it
# for different platforms. For example: c_source.append('windows-specific.c')
c_source = ['code/ibmmq/ibmmqc.c']

ld_flags = []

# Get any platform-specific settings
plat = platform.system()
# Windows
if plat == "Windows":
    library_dirs, include_dirs, libraries = get_windows_settings()

# AIX
elif plat == "AIX":
    library_dirs, include_dirs, libraries = get_aix_settings()

# At this point, to preserve backward-compatibility we try out generic
# UNIX settings first, i.e. libraries and include files in well-known locations.
# Otherwise we look up dspmqver in $PATH.
else:
    has_generic_lib = os.path.exists('/opt/mqm/lib64')
    has_mq_file_path = os.environ.get('MQ_FILE_PATH', False)

    if has_generic_lib or has_mq_file_path:
        library_dirs, include_dirs, libraries = get_generic_unix_settings()

    else:
        # As one more efffort, we try to find
        # the path that dspmqver is installed to and find the rest
        # of the information needed in relation to that base directory.
        dspmqver_path = which('dspmqver')

        # We have found the command so we will be able to extract the relevant directories now
        if dspmqver_path:
            library_dirs, include_dirs, libraries = get_locations_by_command_path(dspmqver_path)
        # Otherwise, just pick the standard Unix directories and hope for the best.
        else:
            library_dirs, include_dirs, libraries = get_generic_unix_settings()

# print('Using library_dirs:`%s`, include:`%s`, libraries:`%s`' % (library_dirs, include_dirs, libraries))

# Can we find the MQ C header files? If not, there's no point in continuing, and we can
# give a reasonable error message immediately instead of trying to decode C compiler errors.
# If we are in the CI environment, we still build the package as we want
# to be able to build the source package without the MQ C headers.
found_headers = False  # pylint: disable=invalid-name
for d in include_dirs:
    p = os.path.join(d, "cmqc.h")
    if os.path.isfile(p):
        found_headers = True
if not found_headers and not os.environ.get('CI', ''):
    msg = "Cannot find MQ C header files.\n"
    msg += "Ensure you have already installed the MQ Client and SDK.\n"
    msg += "Use the MQ_FILE_PATH environment variable to identify a non-default location."
    raise FileNotFoundError(msg)

LONG_DESCRIPTION = """
Python library for IBM MQ
-------------------------

The "ibmmq" package is an open-source Python extension for IBM MQ.

It gives a full-featured implementation of the MQI programming interface, with additional
helper functions to assist with system monitoring and management.

Sample code
-----------

To put a message on a queue:

.. code-block:: python

    import ibmmq

    queue_manager = ibmmq.connect('QM1', 'DEV.APP.SVRCONN', '192.168.1.121(1414)')

    q = ibmmq.Queue(queue_manager, 'DEV.QUEUE.1')
    q.put('Hello from Python!')

To read the message back from the queue:

.. code-block:: python

    import ibmmq

    queue_manager = ibmmq.connect('QM1', 'DEV.APP.SVRCONN', '192.168.1.121(1414)')

    q = ibmmq.Queue(queue_manager, 'DEV.QUEUE.1')
    msg = q.get()
    print('Here is the message:', msg)

Many more examples are in the `project repository
<https://github.com/ibm-messaging/mq-mqi-python/tree/main/code/examples/>`_
and in the `dev-patterns repository
<https://github.com/ibm-messaging/mq-dev-patterns/tree/master/Python/>`_.

"""

# Define how the C module gets built. Set flags to build using the Python 3.9
# Limited API which should make the binary extension forwards compatible.
mqi_extension = Extension('ibmmq.ibmmqc', c_source,
                          define_macros=[('PYVERSION', '"' + VERSION + '"'),
                                         ('Py_LIMITED_API', _ABI_LIMITS['Py_LIMITED_API'])
                                         ],
                          py_limited_api=True,
                          library_dirs=library_dirs,
                          include_dirs=include_dirs,
                          extra_link_args=ld_flags,
                          libraries=libraries)

setup(name='ibmmq',
      version=VERSION,
      description='Python Extension for IBM MQ',
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/x-rst',
      author='IBM MQ Development',
      url='https://ibm.com/software/products/en/ibm-mq',
      download_url='https://github.com/ibm-messaging/mq-mqi-python',
      platforms='OS Independent',
      package_dir={'': 'code'},
      packages=['ibmmq'],
      python_requires=_ABI_LIMITS['python_requires'],
      license_files=['LICENSE*'],
      license='Python-2.0',
      keywords=('pymqi IBMMQ MQ WebSphere WMQ MQSeries IBM middleware messaging queueing asynchronous SOA EAI ESB integration'),
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Natural Language :: English',
                   'Operating System :: OS Independent',
                   'Programming Language :: C',
                   'Programming Language :: Python',
                   'Topic :: Software Development :: Libraries :: Python Modules'],
      ext_modules=[mqi_extension],
      options={"bdist_wheel": {"py_limited_api": _ABI_LIMITS["py_limited_api"]}})



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how to set an option for a channel connection. Though
since these are negotiated with the qmgr, it will only take effect if both
sides agree.
"""

import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
message = 'Hello from Python!' * 10000
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

cd = mq.CD()
cd.MsgCompList[0] = mq.CMQXC.MQCOMPRESS_ZLIBHIGH

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

queue = mq.Queue(qmgr, queue_name)
queue.put(message)
queue.close()

qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how to interrogate the Windows registry to determine
which version of MQ is installed
"""

# Exclude some warnings from pylint that show up when I try to scan this on a Linux machine
# pylint: disable=import-error,undefined-variable
import logging
import _winreg

logging.basicConfig(level=logging.INFO)

# Even though the product has changed names a few times, this key is still maintained
key_name = 'Software\\IBM\\MQSeries\\CurrentVersion'

try:
    key = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, key_name)
except WindowsError:
    logging.info('Could not find IBM MQ-related information in Windows registry.')
else:
    version = _winreg.QueryValueEx(key, 'VRMF')[0]
    logging.info('IBM MQ version is `%s`.', version)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows a request/reply pattern, with the requester waiting for
a reply message that has a CorrelId matching the request's MsgId. This is
one common pattern for linking requests and responses.
The program continues indefinitely, until interrupted from the keyboard.
"""

import logging
import threading
import time
import uuid

import ibmmq as mq

logging.basicConfig(level=logging.INFO)

# Queue manager name
queue_manager = 'QM1'

# Connection host and port
conn_info = '127.0.0.1(1414)'

# Channel to transfer data through
channel = 'DEV.APP.SVRCONN'

# Request Queue
request_queue_name = 'DEV.QUEUE.1'

# ReplyTo Queue
replyto_queue_name = 'DEV.QUEUE.2'

message_prefix = 'Test Data. '

class Producer(threading.Thread):
    """ A base class for any producer used in this example.
    """
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True

        cd = mq.CD()
        cd.ChannelName = channel
        cd.ConnectionName = conn_info
        cd.ChannelType = mq.CMQXC.MQCHT_CLNTCONN
        cd.TransportType = mq.CMQXC.MQXPT_TCP
        self.qm = mq.QueueManager(None)
        self.qm.connect_with_options(queue_manager,
                                     opts=mq.CMQC.MQCNO_HANDLE_SHARE_NO_BLOCK,
                                     cd=cd)

        self.req_queue = mq.Queue(self.qm, request_queue_name)
        self.replyto_queue = mq.Queue(self.qm, replyto_queue_name)


class RequestProducer(Producer):
    """ Instances of this class produce an infinite stream of request messages
    and wait for appropriate responses on reply-to queues.
    """

    def run(self):

        while True:
            # Put the request message.
            put_mqmd = mq.MD()

            # Set the MsgType to request.
            put_mqmd['MsgType'] = mq.CMQC.MQMT_REQUEST

            # Set up the ReplyTo QUeue/Queue Manager (Queue Manager is automatically
            # set by MQ).
            put_mqmd['ReplyToQ'] = replyto_queue_name
            put_mqmd['ReplyToQMgr'] = queue_manager

            # Set up the put options - do it with NO_SYNCPOINT so that the request
            # message is committed immediately.
            put_opts = mq.PMO(Options=mq.CMQC.MQPMO_NO_SYNCPOINT + mq.CMQC.MQPMO_FAIL_IF_QUIESCING)

            # Create a random message.
            message = message_prefix + uuid.uuid4().hex

            self.req_queue.put(message, put_mqmd, put_opts)
            logging.info('Put request message.  Message: [%s]', message)

            # Set up message descriptor for get.
            get_mqmd = mq.MD()

            # Set the get CorrelId to the put MsgId (which was set by MQ on the put1).
            get_mqmd['CorrelId'] = put_mqmd['MsgId']

            # Set up the get options.
            get_opts = mq.GMO(
                Options=mq.CMQC.MQGMO_NO_SYNCPOINT + mq.CMQC.MQGMO_FAIL_IF_QUIESCING + mq.CMQC.MQGMO_WAIT)

            # Version must be set to at least 2 to use the GMO MatchOptions field.
            get_opts['Version'] = mq.CMQC.MQGMO_VERSION_2

            # Tell MQ that we are matching on CorrelId.
            get_opts['MatchOptions'] = mq.CMQC.MQMO_MATCH_CORREL_ID

            # Set the wait timeout of half a second.
            get_opts['WaitInterval'] = 500

            # Open the replyto queue and get response message,
            replyto_queue = mq.Queue(self.qm, replyto_queue_name, mq.CMQC.MQOO_INPUT_SHARED)
            response_message = replyto_queue.get(None, get_mqmd, get_opts)

            logging.info('Got response message `%s`', response_message)

            time.sleep(1)

class ResponseProducer(Producer):
    """ Instances of this class wait for request messages and produce responses.
    """

    def run(self):

        # Request message descriptor, will be reset after processing each
        # request message.
        request_md = mq.MD()

        # Get Message Options
        gmo = mq.GMO()
        gmo.Options = mq.CMQC.MQGMO_WAIT | mq.CMQC.MQGMO_FAIL_IF_QUIESCING
        gmo.WaitInterval = 500  # Half a second

        queue = mq.Queue(self.qm, request_queue_name)

        keep_running = True

        while keep_running:
            try:
                # Wait up to to gmo.WaitInterval for a new message.
                request_message = queue.get(None, request_md, gmo)

                # Create a response message descriptor with the CorrelId
                # set to the value of MsgId of the original request message.
                response_md = mq.MD()
                response_md.CorrelId = request_md.MsgId

                response_message = 'Response to message %s' % request_message
                self.replyto_queue.put(response_message, response_md)

                # Reset the MsgId, CorrelId & GroupId so that we can reuse
                # the same 'md' object again.
                request_md.MsgId = mq.CMQC.MQMI_NONE
                request_md.CorrelId = mq.CMQC.MQCI_NONE
                request_md.GroupId = mq.CMQC.MQGI_NONE

            except mq.MQMIError as e:
                if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_NO_MSG_AVAILABLE:
                    # No messages, that's OK, we can ignore it.
                    pass
                else:
                    # Some other error condition.
                    raise


req = RequestProducer()
resp = ResponseProducer()

req.start()
resp.start()

try:
    while True:
        time.sleep(0.1)
except KeyboardInterrupt:
    req.qm.disconnect()



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows the simplest connection mechanism to a queue manager. Only
the name is required. By default, this will try a local connection.

But setting MQSERVER and either running in an environment where only the MQ client components
are installed or setting MQ_CONNECT_TYPE=CLIENT will try to use a network connection.
"""

import ibmmq as mq

queue_manager = 'QM1'
qmgr = mq.connect(queue_manager)

print("Connection succeeded.")

qmgr.disconnect()



"""MQMDE: Message Descriptor Extension"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class MDE(MQOpts):
    """ Construct an MQMDE Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    This will rarely be needed as the package primarily deals with MDv2 structures. But
    the MQXQH always contains only an MDv1 which in turn might be followed by an MDE that
    has to be parsed.
    """
    def __init__(self, **kw):
        super().__init__(tuple([
            ['_StrucId', CMQC.MQMDE_STRUC_ID, '4s'],
            ['Version', CMQC.MQMDE_VERSION_2, MQLONG_TYPE],
            ['StrucLength', CMQC.MQMDE_LENGTH_2, MQLONG_TYPE],
            ['Encoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
            ['CodedCharSetId', CMQC.MQCCSI_Q_MGR, MQLONG_TYPE],
            ['Format', b'', '8s'],
            ['Flags', CMQC.MQMDEF_NONE, MQLONG_TYPE],
            ['GroupId', b'', '24s'],
            ['MsgSeqNumber', 1, MQLONG_TYPE],
            ['Offset', 0, MQLONG_TYPE],
            ['MsgFlags', CMQC.MQMF_NONE, MQLONG_TYPE],
            ['OriginalLength', CMQC.MQOL_UNDEFINED, MQLONG_TYPE]]), **kw)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows putting several messages to a queue using the ASYNC put option and
then using the MQSTAT operation to see if they all succeeded. The queue is toggled between
PUT(ENABLED) and PUT(DISABLED) to show when and how the error is reported. It is NOT generated
at MQPUT time.
"""

import ibmmq as mq

queue_manager = "QM1"
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)

cd = mq.CD()
cd.ConnectionName = conn_info
cd.ChannelName = channel


cno = mq.CNO()
cno.Options = mq.CMQC.MQCNO_CLIENT_BINDING

csp = mq.CSP()
csp.CSPUserId = 'app'
csp.CSPPassword = 'password'

queue_name = 'DEV.QUEUE.1'
message = 'Hello from Python: message '

qmgr = mq.connect(queue_manager, cd=cd, csp=csp, cno=cno)

od = mq.OD()
od.ObjectName = queue_name
queue = mq.Queue(qmgr, od, mq.CMQC.MQOO_SET | mq.CMQC.MQOO_OUTPUT)

pmo = mq.PMO()
pmo.Options = mq.CMQC.MQPMO_NO_SYNCPOINT | mq.CMQC.MQPMO_ASYNC_RESPONSE

for opt in (mq.CMQC.MQQA_PUT_ALLOWED, mq.CMQC.MQQA_PUT_INHIBITED):
    selectors = {mq.CMQC.MQIA_INHIBIT_PUT: opt}
    queue.set(selectors)
    for i in range(0, 10):
        md = mq.MD()
        queue.put(message + str(i), md, pmo)

    sts = qmgr.stat(mq.CMQC.MQSTAT_TYPE_ASYNC_ERROR)
    print("Queue: ", mq.CMQSTRC.MQQA_PUT_DICT[opt])
    print(f"Put Success: {sts.PutSuccessCount}")
    print(f"Put Failure: {sts.PutFailureCount}")
    print()

# Make sure the queue is back in "allowed" state when we end
selectors = {mq.CMQC.MQIA_INHIBIT_PUT: mq.CMQC.MQQA_PUT_ALLOWED}
queue.set(selectors)

queue.close()
qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how asynchronous consume works. A function is passed
to the MQ library and invoked for message arrival and for certain events
"""

import time
import ibmmq as mq

class Status:
    """A simple way of holding the state without using explicit global variables"""
    ok = True
    connBroken = False


def cb_func(**kwargs):
    """
    The callback function is called with parameters:
       queue_manager, queue, md, gmo, msg, cbc
    The CallbackContext (cbd) structure holds the reasons for invocation, along with
    object and connection-related correlation buffers. Which we print out here

    The function can be defined as a method within a class by adding a "self" parameter.
    """
    msg = kwargs['msg']
    cbc = kwargs['cbc']
    obj = kwargs['queue']
    qmgr = kwargs['queue_manager']

    print(f"In CB Func: QMgr={qmgr.get_name()} Q={obj.get_name() if obj else '<<None>>'} Msg={msg[:cbc.DataLength] if msg else '<<NONE>>'}")
    print(f"    CallType      = {cbc.CallType} [{mq.CMQSTRC.MQCBCT_DICT[cbc.CallType]}]")
    print(f"    CallbackArea  = {cbc.CallbackArea}")
    print(f"    ConnectionArea= {cbc.ConnectionArea}")

    if cbc.Reason != mq.CMQC.MQRC_NONE:
        print(f"Reason: {cbc.Reason} [{mq.CMQSTRC.MQRC_DICT[cbc.Reason]}]")

        # Going to assume for now that any event (other than Reason == 2033) is due to the connection
        # ending. A real program might need to be more specific as this assumption is
        # not always true.
        if cbc.CallType == mq.CMQC.MQCBCT_EVENT_CALL and cbc.Reason != mq.CMQC.MQRC_NO_MSG_AVAILABLE:
            Status.connBroken = True
        else:
            pass
        Status.ok = False


# Define the connection to the queue manager
queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

queue_name = 'DEV.QUEUE.1'
print("Using package version: ", mq.get_versions())

csp = mq.CSP()
csp.CSPUserId = user
csp.CSPPassword = password

qmgr = mq.connect(queue_manager, channel, conn_info, csp=csp)

od = mq.OD()
od.ObjectName = queue_name

q = mq.Queue(qmgr, od, mq.CMQC.MQOO_INPUT_AS_Q_DEF)

# Looking for any message on the queue
md = mq.MD()
gmo = mq.GMO()
gmo.Options = mq.CMQC.MQGMO_NO_SYNCPOINT
gmo.Options |= mq.CMQC.MQGMO_WAIT
gmo.WaitInterval = 3 * 1000  # convert to milliseconds

# Register a callback for the queue
cbd = mq.CBD()
cbd.CallbackFunction = cb_func
cbd.CallbackArea = b'My CB Area'
cbd.CallbackType = mq.CMQC.MQCBT_MESSAGE_CONSUMER
q.cb(operation=mq.CMQC.MQOP_REGISTER, cbd=cbd, md=md, gmo=gmo)

# Also register a callback for qmgr-wide events
cbd.CallbackFunction = cb_func
cbd.CallbackType = mq.CMQC.MQCBT_EVENT_HANDLER
cbd.CallbackArea = b'MY QGMGR CB AREA'
qmgr.cb(operation=mq.CMQC.MQOP_REGISTER, cbd=cbd)

# Start the async processing
ctlo = mq.CTLO()
ctlo.ConnectionArea = b'My Connection Area'
qmgr.ctl(mq.CMQC.MQOP_START, ctlo)

# Now wait for messages to be put to the queue or for qmgr events
# such as endmqm
while Status.ok:
    print("Sleeping ...")
    time.sleep(2)

if not Status.connBroken:
    qmgr.ctl(mq.CMQC.MQOP_STOP, ctlo)
    q.cb(operation=mq.CMQC.MQOP_DEREGISTER)

    q.close()
    qmgr.disconnect()

print("Done.")



"""Classes for executing PCF commands
"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, CMQCFC, CMQSTRC, MD, PMO, GMO, OD
from mqqueue import Queue
from mqpcf import *

from mqqmgr import *

# pylint: disable=no-member

try:
    from typing import Any, Union, Dict
except ImportError:
    pass

class ByteString:
    """ A simple wrapper around string values, suitable for passing into
    calls wherever IBM's docs state a 'byte string' object should be passed in.
    """
    def __init__(self, value: bytes):
        self.value = value
        self._byte_string = True

    def __len__(self) -> int:
        return len(self.value)

class _Filter:
    """ The base class for PCF filters. The initializer expects the user to provide
    the selector, value and the operator to use. For instance, they can be respectively
    MQCA_Q_DESC, 'MY.QUEUE.*', MQCFOP_LIKE. Compare with the public Filter class.
    """
    _filter_type = None  # type: Union[str, None]

    def __init__(self, selector, value, operator):
        self.selector = selector  # this is int
        self.value = ensure_strings_are_bytes(value)
        self.operator = operator  # this is int

    def __repr__(self):
        msg = '<%s at %s %s:%s:%s>'
        return msg % (self.__class__.__name__, hex(id(self)), self.selector, self.value, self.operator)

# ################################################################################################################################

class StringFilter(_Filter):
    """ A subclass of _Filter suitable for passing string filters around.
    """
    _filter_type = 'string'

# ################################################################################################################################

class IntegerFilter(_Filter):
    """ A subclass of _Filter suitable for passing integer filters around.
    """
    _filter_type = 'integer'

# ################################################################################################################################

class ByteStringFilter(_Filter):
    """ A subclass of _Filter suitable for passing byte string filters around.
    """
    _filter_type = 'bytestring'

###############################################################################################################################

class FilterOperator:
    """ Creates low-level filters based on what's been provided in the high-level
    Filter object.
    """
    operator_mapping = {'less': CMQCFC.MQCFOP_LESS,
                        'equal': CMQCFC.MQCFOP_EQUAL,
                        'greater': CMQCFC.MQCFOP_GREATER,
                        'not_less': CMQCFC.MQCFOP_NOT_LESS,
                        'not_equal': CMQCFC.MQCFOP_NOT_EQUAL,
                        'not_greater': CMQCFC.MQCFOP_NOT_GREATER,
                        'like': CMQCFC.MQCFOP_LIKE,
                        'not_like': CMQCFC.MQCFOP_NOT_LIKE,
                        'contains': CMQCFC.MQCFOP_CONTAINS,
                        'excludes': CMQCFC.MQCFOP_EXCLUDES,
                        'contains_gen': CMQCFC.MQCFOP_CONTAINS_GEN,
                        'excludes_gen': CMQCFC.MQCFOP_EXCLUDES_GEN,
                        }  # type: Dict[str, int]

    def __init__(self, selector: int, operator_name: str):
        self.filter_cls = _Filter  # Start with the parent class as the first ref is how mypy sees it
        if CMQC.MQIA_FIRST <= selector <= CMQC.MQIA_LAST:
            self.filter_cls = IntegerFilter
        elif CMQC.MQCA_FIRST <= selector <= CMQC.MQCA_LAST:
            self.filter_cls = StringFilter
        elif CMQC.MQBA_FIRST <= selector <= CMQC.MQBA_LAST:
            self.filter_cls = ByteStringFilter
        else:
            msg = 'selector [%s] is of an unsupported type (not an integer ' + \
                'or string or bytestring attribute).'
            raise Error(msg % selector)
        self.selector = selector
        self.operator = self.operator_mapping.get(operator_name)
        # Do we support the operator?
        if not self.operator:
            msg = 'Operator [%s] is not supported.'
            raise Error(msg % operator_name)

    def __call__(self, value: Union[str, int]) -> _Filter:  # Union[IntegerFilter, StringFilter, ByteStringFilter]:
        ensure_not_unicode(value)
        return (self.filter_cls)(self.selector, value, self.operator)

# ################################################################################################################################

class Filter:
    """ The user-facing filtering class which provides syntactic sugar
    on top of _Filter and its subclasses.
    """
    def __init__(self, selector: int):
        self.selector = selector

    def __getattribute__(self, name: str) -> FilterOperator:
        """ A generic method for either fetching the Filter object's
        attributes or calling magic methods like 'like', 'contains' etc.
        """
        if name == 'selector':
            return object.__getattribute__(self, name)

        return FilterOperator(self.selector, name)

# ################################################################################################################################

#
# This piece of magic shamelessly plagiarised from xmlrpclib.py. It
# works a bit like a C++ STL functor.
#
class _Method:
    def __init__(self, pcf, name):
        # type: (PCFExecute, str) -> None
        self.__pcf = pcf
        self.__name = name

    def __getattr__(self, name):
        # type: (str) -> _Method
        return _Method(self.__pcf, '%s.%s' % (self.__name, name))

    def __call__(self, *args):
        # type: (Union[dict, list, _Filter]) -> list
        mqlog.trace_entry("admin:method:__call__")

        if self.__name[0:7] == 'CMQCFC.':
            self.__name = self.__name[7:]
        if self.__pcf.qm:
            bytes_encoding = self.__pcf.bytes_encoding
            _ = self.__pcf.qm.getHandle()
        else:
            bytes_encoding = 'utf8'
            _ = self.__pcf.getHandle()

        len_args = len(args)

        if len_args == 2:
            args_dict, filters = args

        elif len_args == 1:
            args_dict, filters = args[0], []

        else:
            args_dict, filters = {}, []

        # MQCFH_VERSION_3 means we have a minimum of MQ V6 as the
        # target qmgr.
        mqcfh = CFH(Version=CMQCFC.MQCFH_VERSION_3,
                    Command=CMQCFC.__dict__[self.__name],
                    Type=CMQCFC.MQCFT_COMMAND_XR,
                    ParameterCount=len(args_dict) + len(filters))
        message = mqcfh.pack()

        parameter = MQOpts([])  # Set an initial type, to satisfy mypy

        if args_dict:
            if isinstance(args_dict, dict):
                for key, value in args_dict.items():
                    if isinstance(value, (str, bytes)):
                        if is_unicode(value):
                            value = value.encode(bytes_encoding)

                        parameter = CFST(Parameter=key,
                                         String=value)
                    elif isinstance(value, ByteString):
                        parameter = CFBS(Parameter=key,
                                         String=value.value)
                    elif isinstance(value, int):
                        # Backward compatibility for older behaviour
                        # returning a single value instead of a list
                        is_list = False
                        for item in CMQCFC.__dict__:
                            if ((item[:7] == 'MQIACF_' or item[:7] == 'MQIACH_')
                               and item[-6:] == '_ATTRS'
                               and CMQCFC.__dict__[item] == key):

                                is_list = True
                                break

                        if not is_list:
                            parameter = CFIN(Parameter=key, Value=value)
                        else:
                            parameter = CFIL(Parameter=key, Values=[value])
                    elif isinstance(value, list):
                        if isinstance(value[0], int):
                            parameter = CFIL(Parameter=key, Values=value)
                        elif isinstance(value[0], (str, bytes)):
                            _value = []
                            for item in value:
                                if is_unicode(item):
                                    item = item.encode(bytes_encoding)
                                _value.append(item)
                            value = _value

                            parameter = CFSL(Parameter=key, Strings=value)

                    message = message + parameter.pack()
            elif isinstance(args_dict, list):
                for parameter in args_dict:
                    message = message + parameter.pack()

        if filters:
            for pcf_filter in filters:
                if isinstance(pcf_filter, _Filter):
                    if pcf_filter._filter_type == 'string':
                        pcf_filter = CFSF(Parameter=pcf_filter.selector,
                                          Operator=pcf_filter.operator,
                                          FilterValue=pcf_filter.value)
                    elif pcf_filter._filter_type == 'integer':
                        pcf_filter = CFIF(Parameter=pcf_filter.selector,
                                          Operator=pcf_filter.operator,
                                          FilterValue=pcf_filter.value)
                    elif pcf_filter._filter_type == 'bytestring':
                        pcf_filter = CFBF(Parameter=pcf_filter.selector,
                                          Operator=pcf_filter.operator,
                                          FilterValue=pcf_filter.value)

                message = message + pcf_filter.pack()

        # Either open the command queue, or use the pre-opened Queue object
        command_queue_opened = False
        if self.__pcf._command_queue is None:
            assert self.__pcf.qm is not None  # Keep mypy happy
            command_queue = Queue(self.__pcf.qm,
                                  self.__pcf.command_queue_name,
                                  CMQC.MQOO_OUTPUT)
            self.__pcf._command_queue = command_queue
            command_queue_opened = True
        else:
            command_queue = self.__pcf._command_queue

        if self.__pcf.command_timeout == 0:
            # If command timout was not specified,
            # multiply the wait time to give a reasonable expiry time
            # Convert to tenths and then x3 (for out, back and delay).
            expiry = (self.__pcf.response_wait_interval // 100) * 3
        else:
            # Use explicitly specified value, converted to tenths of seconds
            expiry = self.__pcf.command_timeout // 100

        put_md = MD(Format=CMQC.MQFMT_ADMIN,
                    MsgType=CMQC.MQMT_REQUEST,
                    ReplyToQ=self.__pcf.reply_queue_name,
                    Feedback=CMQC.MQFB_NONE,
                    Expiry=expiry,
                    Report=CMQC.MQRO_PASS_DISCARD_AND_EXPIRY | CMQC.MQRO_DISCARD_MSG)
        put_opts = PMO(Options=CMQC.MQPMO_NO_SYNCPOINT)

        command_queue.put(message, put_md, put_opts)

        # If it was opened for this run, then close it again immediately
        if command_queue_opened:
            command_queue.close()
            self.__pcf._command_queue = None

        gmo_options = (CMQC.MQGMO_NO_SYNCPOINT |
                       CMQC.MQGMO_FAIL_IF_QUIESCING |
                       CMQC.MQGMO_WAIT)

        if self.__pcf.convert:
            gmo_options |= CMQC.MQGMO_CONVERT

        get_opts = GMO(
            Options=gmo_options,
            Version=CMQC.MQGMO_VERSION_2,
            MatchOptions=CMQC.MQMO_MATCH_CORREL_ID,
            WaitInterval=self.__pcf.response_wait_interval)
        get_md = MD(CorrelId=put_md.MsgId)

        # Initialise the list of responses
        ress = []

        got_all_replies = False
        in_cmdscope = False

        while True:
            try:
                message = self.__pcf.reply_queue.get(None, get_md, get_opts)
                res, mqcfh_response = self.__pcf.unpack(message)

                t = mqcfh_response.Type

                # The XR messages come back from a z/OS queue manager. For an individual qmgr operation
                # they are essentially meaningless. But for something with a CMDSCOPE flag, then they
                # show how we need to process and wait for all responses.
                # Note: The MQCACF_RESPONSE_Q_MGR_NAME element shows which qmgr replied when using CMDSCOPE
                flag = 0
                if t == CMQCFC.MQCFT_XR_MSG:
                    # Look for flag saying cmdscope ACCEPTED
                    try:
                        flag = res.get(CMQCFC.MQIACF_COMMAND_INFO)
                        if flag == CMQCFC.MQCMDI_CMDSCOPE_ACCEPTED:
                            in_cmdscope = True
                    except KeyError:
                        pass

                # Have all the replies been received? Different for CMDSCOPE(*) compared to regular
                if in_cmdscope:
                    if t == CMQCFC.MQCFT_XR_MSG:
                        try:
                            flag = res.get(CMQCFC.MQIACF_COMMAND_INFO)
                            if flag == CMQCFC.MQCMDI_CMDSCOPE_COMPLETED:
                                got_all_replies = True
                        except KeyError:
                            pass
                elif mqcfh_response.Control == CMQCFC.MQCFC_LAST:
                    got_all_replies = True

                # Don't add these to the responses
                if t in (CMQCFC.MQCFT_XR_MSG, CMQCFC.MQCFT_XR_SUMMARY):
                    res = None

                if res:
                    ress.append(res)

                # print(f"Flag: {flag} incmd: {in_cmdscope} gotall: {got_all_replies}")
                if got_all_replies:
                    break

            except MQMIError as e:
                # There might be something special we want to do with 2033s.
                # But for now, just report it in the same way as any other failure.
                mqlog.trace_exit("admin:method:__call__", ep=1, rc=e.reason)

                if e.reason == CMQC.MQRC_NO_MSG_AVAILABLE:
                    # print("Timed out...")
                    raise e
                    # return ress
                raise e
        mqlog.trace_exit("admin:method:__call__")

        return ress

# ################################################################################################################################

def _merge_dicts(*dict_args):
    result = {}
    for d in dict_args:
        result.update(d)
    return result

#
# Execute a PCF commmand. Inspired by Maas-Maarten Zeeman
#
class PCFExecute(QueueManager):

    """Send PCF commands or inquiries. PCFExecute must be connected to the Queue Manager
    (using one of the techniques inherited from QueueManager) before
    it's used. Commands are executed by calling a CMQC defined
    MQCMD_* method on the object.  """

    qm = None  # type: Union[QueueManager, None]

    iaStringDict = _merge_dicts(CMQSTRC.MQIA_DICT, CMQSTRC.MQIACF_DICT, CMQSTRC.MQIACH_DICT)
    caStringDict = _merge_dicts(CMQSTRC.MQCA_DICT, CMQSTRC.MQCACF_DICT, CMQSTRC.MQCACH_DICT)
    baStringDict = _merge_dicts(CMQSTRC.MQBACF_DICT)

    def __init__(self, name=None,
                 model_queue_name=b'SYSTEM.DEFAULT.MODEL.QUEUE',
                 reply_queue_name=None,
                 reply_queue_open_options=CMQC.MQOO_INPUT_EXCLUSIVE,  # Allows you to override and set, say, MQOO_READ_AHEAD
                 dynamic_queue_name=b'PYMQPCF.*',
                 command_queue_name=b'SYSTEM.ADMIN.COMMAND.QUEUE',
                 command_queue=None,
                 response_wait_interval=5000,  # 5 seconds
                 command_timeout=0,            # If 0, calculate based on wait interval
                 convert=True):
        # type: (Any, Union[str,bytes], Union[None,bytes,str], int, Union[str,bytes], Union[str,bytes], Union[None,Queue], int, int, bool) -> None
        """PCFExecute(name = '')

        Connect to the Queue Manager 'name' (default value '') ready
        for a PCF command. If name is a QueueManager instance, it is
        used for the connection, otherwise a new connection is made """

        mqlog.trace_entry("admin:pcfexecute:__init__")

        # The default value gives backwards-compatible version where the
        # command queue is opened for each execution. But we can supply
        # a pre-opened queue handle
        self._command_queue = command_queue

        self.__command_queue_name = command_queue_name

        self.__convert = convert

        # Don't allow Unlimited waits. Force it to something large, but
        # not forever. This gives an hour. Setting it to zero would also be foolish,
        # as there would be no time at all to wait for the response. But we'll allow you to
        # find that for yourself.
        if response_wait_interval < 0:
            response_wait_interval = 60 * 60 * 1000
        self.__response_wait_interval = response_wait_interval
        self.__command_timeout = command_timeout

        if model_queue_name and reply_queue_name:
            mqlog.trace_exit("admin:pcfexecute:__init__", ep=1)
            raise PYIFError('Do not specify both a model_queue_name and a reply_queue_name')

        # From here, we can treat the 2 qnames as equivalent. So assign to one and use it.
        # There is an internal __reply_queue_name field, but that comes from the result of the
        # MQOPEN - gives the TDQ name if that's been created.
        if reply_queue_name:
            model_queue_name = reply_queue_name

        if isinstance(name, QueueManager):
            self.qm = name
            super().__init__(None)
        else:
            super().__init__(name)
            self.qm = self

        od = OD(ObjectName=model_queue_name,
                DynamicQName=dynamic_queue_name)

        # Do the open options include any of the INPUT options? If not, then add in one of them.
        open_input_mask = CMQC.MQOO_INPUT_EXCLUSIVE | CMQC.MQOO_INPUT_AS_Q_DEF | CMQC.MQOO_INPUT_SHARED
        if reply_queue_open_options & open_input_mask == 0:
            reply_queue_open_options |= CMQC.MQOO_INPUT_EXCLUSIVE

        self.__reply_queue = Queue(self.qm, od, reply_queue_open_options)
        self.__reply_queue_name = od.ObjectName.strip()
        mqlog.trace_exit("admin:pcfexecute:__init__")

    @property
    def command_queue_name(self):
        """Return the command queue name"""
        return self.__command_queue_name

    @property
    def convert(self):
        """Return whether data conversion is active"""
        return self.__convert

    @property
    def reply_queue(self):
        """Return the reply queue object"""
        return self.__reply_queue

    @property
    def reply_queue_name(self):
        """Return the reply queue name"""
        return self.__reply_queue_name

    @property
    def response_wait_interval(self):
        """Return the wait interval for this object"""
        return self.__response_wait_interval

    @property
    def command_timeout(self):
        """Return the command timeout for this object"""
        return self.__command_timeout

    def __getattr__(self, name):
        """MQCMD_*(attrDict)

        Execute the PCF command or inquiry, passing an an optional
        dictionary of MQ attributes.  The command must begin with
        MQCMD_, and must be one of those defined in the CMQC
        module. If attrDict is passed, its keys must be an attribute
        as defined in the CMQC or CMQC modules (MQCA_*, MQIA_*,
        MQCACH_* or MQIACH_*). The key value must be an int or string,
        as appropriate.

        If an inquiry is executed, a list of dictionaries (one per
        matching query) is returned. Each dictionary encodes the
        attributes and values of the object queried. The keys are as
        defined in the CMQC module (MQIA_*, MQCA_*), The values are
        strings or ints, as appropriate.

        If a command was executed, or no inquiry results are
        available, an empty list is returned.  """

        return _Method(self, name)

    @staticmethod
    def stringify_keys(raw_dict):
        """stringifyKeys(raw_dict)

        Return raw_dict with its keys converted to string
        mnemonics, as defined in CMQC. """

        rv = {}
        for k in raw_dict.keys():
            if isinstance(raw_dict[k], bytes):
                d = PCFExecute.caStringDict
            elif isinstance(raw_dict[k], str):
                raise TypeError('In Python 3 use bytes, not str (found "{0}":"{1}")'.format(k, raw_dict[k]))
            else:
                d = PCFExecute.iaStringDict
            try:
                rv[d[k]] = raw_dict[k]
            except KeyError:
                rv[k] = raw_dict[k]
        return rv

    # Backward compatibility
    stringifyKeys = stringify_keys

    def disconnect(self):
        """ Disconnect from reply_queue
        """
        mqlog.trace_entry("admin:pcfexecute:disconnect")

        try:
            if self.__reply_queue and self.__reply_queue.get_handle():
                self.__reply_queue.close()
        except MQMIError:
            pass
        finally:
            self.__reply_queue = None
            self.__reply_queue_name = None

        mqlog.trace_exit("admin:pcfexecute:disconnect")

    @staticmethod
    def unpack(message: bytes) -> tuple[dict, CFH]:
        """Unpack PCF message to dictionary
        """

        mqcfh = CFH(Version=CMQCFC.MQCFH_VERSION_1)
        mqcfh.unpack(message[:CMQCFC.MQCFH_STRUC_LENGTH])

        if mqcfh.Version != CMQCFC.MQCFH_VERSION_1:
            mqcfh = CFH(Version=mqcfh.Version)
            mqcfh.unpack(message[:CMQCFC.MQCFH_STRUC_LENGTH])

        if mqcfh.CompCode != CMQC.MQCC_OK:
            raise MQMIError(mqcfh.CompCode, mqcfh.Reason)

        res = {}  # type: Dict[str, Union[int, str, bool, Dict]]
        index = mqcfh.ParameterCount
        cursor = CMQCFC.MQCFH_STRUC_LENGTH
        parameter = None  # type: Union[MQOpts, None]
        group = None  # type: Union[None, Dict[str, Union[str, int, bool]]]
        group_count = 0

        while index > 0:

            value = None  # Will always be set by one of these clauses
            parameter_type = struct.unpack(MQLONG_TYPE, message[cursor:cursor + 4])[0]

            if group_count == 0:
                group = None
            if group is not None:
                group_count -= 1
            if parameter_type == CMQCFC.MQCFT_STRING:
                parameter = CFST()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFST_STRUC_LENGTH_FIXED])
                if parameter.StringLength > 1:
                    parameter = CFST(StringLength=parameter.StringLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                # The parameter.String contents might include padding bytes that round up the length.
                # Ideally we'd truncate, but this was behaviour in the original library so I'm reluctant
                # to fix it.
                value = parameter.String   # [:parameter.StringLength] - would truncate
            elif parameter_type == CMQCFC.MQCFT_STRING_LIST:
                parameter = CFSL()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFSL_STRUC_LENGTH_FIXED])
                if parameter.StringLength > 1:
                    parameter = CFSL(StringLength=parameter.StringLength,
                                     Count=parameter.Count,
                                     StrucLength=parameter.StrucLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                # CFSL doesn't have the same padding issue as CFST as all supported
                # attributes are multiples of 4.
                value = parameter.Strings
            elif parameter_type == CMQCFC.MQCFT_INTEGER:
                parameter = CFIN()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFIN_STRUC_LENGTH])
                value = parameter.Value
            elif parameter_type == CMQCFC.MQCFT_INTEGER64:
                parameter = CFIN64()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFIN64_STRUC_LENGTH])
                value = parameter.Value
            elif parameter_type == CMQCFC.MQCFT_INTEGER_LIST:
                parameter = CFIL()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFIL_STRUC_LENGTH_FIXED])
                if parameter.Count > 0:
                    parameter = CFIL(Count=parameter.Count,
                                     StrucLength=parameter.StrucLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                value = parameter.Values
            elif parameter_type == CMQCFC.MQCFT_INTEGER64_LIST:
                parameter = CFIL64()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFIL64_STRUC_LENGTH_FIXED])
                if parameter.Count > 0:
                    parameter = CFIL64(Count=parameter.Count,
                                       StrucLength=parameter.StrucLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                value = parameter.Values
            elif parameter_type == CMQCFC.MQCFT_GROUP:
                parameter = CFGR()
                parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                group_count = parameter.ParameterCount
                index += group_count
                group = {}
                res[parameter.Parameter] = res.get(parameter.Parameter, [])
                res[parameter.Parameter].append(group)
            elif parameter_type == CMQCFC.MQCFT_BYTE_STRING:
                parameter = CFBS()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFBS_STRUC_LENGTH_FIXED])
                if parameter.StringLength > 1:
                    parameter = CFBS(StringLength=parameter.StringLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                value = parameter.String
            elif parameter_type == CMQCFC.MQCFT_STRING_FILTER:
                parameter = CFSF()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFSF_STRUC_LENGTH_FIXED])
                if parameter.FilterValueLength > 0:
                    parameter = CFSF(FilterValueLength=parameter.FilterValueLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                value = (parameter.Operator, parameter.FilterValue)
            elif parameter_type == CMQCFC.MQCFT_BYTE_STRING_FILTER:
                parameter = CFBF()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFBF_STRUC_LENGTH_FIXED])
                if parameter.FilterValueLength > 0:
                    parameter = CFBF(FilterValueLength=parameter.FilterValueLength)
                    parameter.unpack(message[cursor:cursor + parameter.StrucLength])
                value = (parameter.Operator, parameter.FilterValue)
            elif parameter_type == CMQCFC.MQCFT_INTEGER_FILTER:
                parameter = CFIF()
                parameter.unpack(message[cursor:cursor + CMQCFC.MQCFIF_STRUC_LENGTH])
                value = (parameter.Operator, parameter.FilterValue)
            else:
                pcf_type = struct.unpack(MQLONG_TYPE, message[cursor:cursor + 4])
                raise NotImplementedError('Unpack for type ({}) not implemented'.format(pcf_type))
            index -= 1
            cursor += parameter.StrucLength
            if parameter.Type == CMQCFC.MQCFT_GROUP:
                continue
            if group is not None:
                group[parameter.Parameter] = value
            else:
                res[parameter.Parameter] = value

        return res, mqcfh



""" Tests for mq.QueueManager class.
"""

# pylint: disable=missing-function-docstring,no-name-in-module

import unittest

import ibmmq as mq

import config
import utils

try:
    from typing import List
except ImportError:
    pass

from test_setup import Tests
from test_setup import main

class TestQueueManager(Tests):
    """Testing the queue manager"""
    CHCKCLNT = None  # type: int

    @classmethod
    def setUpClass(cls):
        """Initialize test environment."""
        super(TestQueueManager, cls).setUpClass()

        # Get CHCKCLNT value
        attrs = []  # type: List[mq.MQOpts]
        attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_Q_MGR_ATTRS,
                                Values=[mq.CMQC.MQCA_CONN_AUTH]))
        results = cls.pcf.MQCMD_INQUIRE_Q_MGR(attrs)
        conn_auth_name = results[0][mq.CMQC.MQCA_CONN_AUTH]

        attrs = []  # type: List[mq.MQOpts]

        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_AUTH_INFO_NAME,
                                String=conn_auth_name))
        attrs.append(mq.CFIN(Parameter=mq.CMQC.MQIA_AUTH_INFO_TYPE,
                                Value=mq.CMQC.MQAIT_IDPW_OS))
        attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_AUTH_INFO_ATTRS,
                                Values=[mq.CMQC.MQIA_CHECK_CLIENT_BINDING]))

        results = cls.pcf.MQCMD_INQUIRE_AUTH_INFO(attrs)
        cls.CHCKCLNT = results[0][mq.CMQC.MQIA_CHECK_CLIENT_BINDING]

        # Add required rights for pinging QMGR
        attrs = []  # type: List[mq.MQOpts]
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACF_AUTH_PROFILE_NAME,
                                String=b'SYSTEM.DEFAULT.MODEL.QUEUE'))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_OBJECT_TYPE,
                                Value=mq.CMQC.MQOT_Q))
        attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_AUTH_ADD_AUTHS,
                                Values=[mq.CMQCFC.MQAUTH_DISPLAY,
                                        mq.CMQCFC.MQAUTH_INPUT]))
        attrs.append(mq.CFSL(Parameter=mq.CMQCFC.MQCACF_PRINCIPAL_ENTITY_NAMES,
                                Strings=[utils.py3str2bytes(cls.app_user)]))
        cls.pcf.MQCMD_SET_AUTH_REC(attrs)

        attrs = []  # type: List[mq.MQOpts]
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACF_AUTH_PROFILE_NAME,
                                String=b'SYSTEM.ADMIN.COMMAND.QUEUE'))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_OBJECT_TYPE,
                                Value=mq.CMQC.MQOT_Q))
        attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_AUTH_ADD_AUTHS,
                                Values=[mq.CMQCFC.MQAUTH_OUTPUT]))
        attrs.append(mq.CFSL(Parameter=mq.CMQCFC.MQCACF_PRINCIPAL_ENTITY_NAMES,
                                Strings=[utils.py3str2bytes(cls.app_user)]))
        cls.pcf.MQCMD_SET_AUTH_REC(attrs)

        attrs = []  # type: List[mq.MQOpts]
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACF_AUTH_PROFILE_NAME,
                                String=utils.py3str2bytes(cls.queue_manager)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_OBJECT_TYPE,
                                Value=mq.CMQC.MQOT_Q_MGR))
        attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_AUTH_ADD_AUTHS,
                                Values=[mq.CMQCFC.MQAUTH_DISPLAY]))
        attrs.append(mq.CFSL(Parameter=mq.CMQCFC.MQCACF_PRINCIPAL_ENTITY_NAMES,
                                Strings=[utils.py3str2bytes(cls.app_user)]))
        results = cls.pcf.MQCMD_SET_AUTH_REC(attrs)

        if cls.pcf.is_connected:
            cls.pcf.disconnect()
        if cls.qmgr.is_connected:
            cls.qmgr.disconnect()

    def test_init_none(self):
        qmgr = mq.QueueManager(None)
        self.assertFalse(qmgr.is_connected)

    @utils.with_env_complement('MQSERVER', config.MQ.APP_MQSERVER)
    def test_init_name(self):
        # As the connect method provides no way to supply user & password, this
        # cannot work if the queue manager requires it
        if (config.MQ.QM.CONNAUTH.USE_PW == 'REQUIRED') or (self.CHCKCLNT == mq.CMQCFC.MQCHK_REQUIRED):
            self.skipTest('Test not viable for user/password-requiring queue manager')
            return

        # connecting with queue manager name needs MQSERVER set properly
        qmgr = mq.QueueManager(self.queue_manager)
        self.assertTrue(qmgr.is_connected)

        if qmgr.is_connected:
            qmgr.disconnect()

    @utils.with_env_complement('MQSERVER', config.MQ.APP_MQSERVER)
    def test_connect(self):
        # As the connect method provides no way to supply user & password, this
        # cannot work if the queue manager requires it
        if (config.MQ.QM.CONNAUTH.USE_PW == 'REQUIRED') or (self.CHCKCLNT == mq.CMQCFC.MQCHK_REQUIRED):
            self.skipTest('Test not viable for user/password-requiring queue manager')
            return

        qmgr = mq.QueueManager(None)
        self.assertFalse(qmgr.is_connected)
        qmgr.connect(self.queue_manager)
        self.assertTrue(qmgr.is_connected)
        if qmgr.is_connected:
            qmgr.disconnect()

    def test_connect_tcp_client(self):
        qmgr = mq.QueueManager(None)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.channel, self.conn_info, user=self.user,
            password=self.password)
        self.assertTrue(qmgr.is_connected)
        if qmgr.is_connected:
            qmgr.disconnect()

    def test_connect_tcp_client_without_cred(self):
        if (config.MQ.QM.CONNAUTH.USE_PW == 'REQUIRED') or (self.CHCKCLNT == mq.CMQCFC.MQCHK_REQUIRED):
            self.skipTest('Test not viable for user/password-requiring queue manager')
            return

        qmgr = mq.QueueManager(None)
        with self.assertRaises(mq.MQMIError) as ex_ctx:
            qmgr.connect_tcp_client(
                self.queue_manager, mq.cd(), self.channel, self.conn_info)
            self.assertEqual(ex_ctx.exception.reason, mq.CMQC.MQRC_NOT_AUTHORIZED)
        if qmgr.is_connected:
            qmgr.disconnect()

    # This test should be run last
    # ConnectionName list with unaccessible QM affects on channel name of the next test if MQSERVER used
    # changing the order of ConnectionName entries does not affect to issue occurrence
    def test_zzz_connect_tcp_client_conection_list(self):
        qmgr = mq.QueueManager(None)
        conn_info = '127.0.0.1(22),{0}'.format(self.conn_info)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.channel, conn_info, user=self.user,
            password=self.password)
        self.assertTrue(qmgr.is_connected)
        if qmgr.is_connected:
            qmgr.disconnect()

    # This test overlaps with test_mq80.test_successful_connect_without_optional_credentials,
    # but hey, why not
    def test_connect_tcp_client_with_none_credentials(self):
        if config.MQ.QM.CONNAUTH.USE_PW == 'REQUIRED' or self.CHCKCLNT == mq.CMQCFC.MQCHK_REQUIRED:
            self.skipTest('Test not viable for user/password-requiring queue manager')
            return

        qmgr = mq.QueueManager(None)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.app_channel, self.conn_info, user=None,
            password=None)
        self.assertTrue(qmgr.is_connected)
        if qmgr.is_connected:
            qmgr.disconnect()

    def test_disconnect(self):
        qmgr = mq.QueueManager(None)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.channel, self.conn_info, user=self.user,
            password=self.password)
        self.assertTrue(qmgr.is_connected)
        if qmgr.is_connected:
            qmgr.disconnect()
            self.assertFalse(qmgr.is_connected)

    def test_get_handle_unconnected(self):
        qmgr = mq.QueueManager(None)
        self.assertRaises(mq.PYIFError, qmgr.get_handle)

    def test_get_handle_connected(self):
        qmgr = mq.QueueManager(None)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.channel, self.conn_info, user=self.user,
            password=self.password)
        handle = qmgr.get_handle()
        self.assertTrue(isinstance(handle, int))

        if qmgr.is_connected:
            qmgr.disconnect()

    @unittest.skip('Not implemented yet')
    def test_begin(self):
        pass

    @unittest.skip('Not implemented yet')
    def test_commit(self):
        pass

    @unittest.skip('Not implemented yet')
    def test_backout(self):
        pass

    def test_inquire(self):
        qmgr = mq.QueueManager(None)
        qmgr.connect_tcp_client(
            self.queue_manager, mq.cd(), self.channel, self.conn_info, user=self.user,
            password=self.password)
        attribute = mq.CMQC.MQCA_Q_MGR_NAME
        expected_value = utils.py3str2bytes(self.queue_manager)
        attribute_value = qmgr.inquire(attribute)
        self.assertEqual(len(attribute_value), mq.CMQC.MQ_Q_MGR_NAME_LENGTH)
        self.assertTrue(utils.strcmp(attribute_value.strip(), expected_value))

        if qmgr.is_connected:
            qmgr.disconnect()
            self.assertFalse(qmgr.is_connected)


if __name__ == '__main__':
    main(module="test_queue_manager")



"""This class is intended to be superclass for other MQ Object types.
It is a fairly empty class to start with. Can migrate values like the
object handle from subclasses, and functions such as open(), close() or inq() later.
"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.
from mqcommon import *

class MQObject:
    """
    Note that the name might not be available at construction time; the subclass
    will have to set it when known. This will allow us to potentially create classes (or
    use this one directly) for Namelist and Process objects that can be worked with in the MQI.
    """
    def __init__(self, name):
        # print(f"In MQObject constructor for class {type(self)} with value {name}")
        self._name = name

    def to_string(self, v, encoding=EncodingDefault.bytes_encoding):
        """Use the specified encoding to convert MQCHAR[] to a Python3 string, stripping trailing NULs/spaces.
        If there's an error, return the input unchanged.
        """
        if isinstance(v, bytes):
            try:
                null_index = v.find(0)
                if null_index != -1:
                    v = v[:null_index]
                return v.decode(encoding).strip()
            except UnicodeError:
                pass
        return v



"""Test the PubSub verbs"""

import sys
import unicodedata
import unittest
import ibmmq as mq
import utils
import config

if sys.version_info[0] >= 3:
    def unicode(x, encoding):
        """Turn Python3 strings into bytes"""
        if isinstance(x, bytes):
            return x.decode(encoding)
        return str(x)  # In py 3 every string is unicode.

# pylint: disable=missing-function-docstring

class TestPubSub(unittest.TestCase):  # pylint: disable=too-many-instance-attributes

    """Test Pub/Sub with the following six test cases:
    |-------+---------+-------------+----------+-------------|
    |       | Managed | Managed     | Provided | Provided    |
    |       | Durable | Non-Durable | Durable  | Non-Durable |
    |-------+---------+-------------+----------+-------------|
    | Admin | x       |             | x        |             |
    | API   | x       | x           | x        | x           |
    |-------+---------+-------------+----------+-------------|
    The combination Admin/Non-Durable is not possible.
    All tests follow the same procedure:
    1. Create or register the subscription
    2. Publish a message
    3. get() the message
    4. compare result with message
    5. tearDown(): delete objects
    """

    def cleanup(self):
        pcf = mq.PCFExecute(self.qmgr)
        try:
            qprefix = config.MQ.QUEUE.PREFIX + "*"
            args = {mq.CMQC.MQCA_Q_NAME: qprefix}
            response = pcf.MQCMD_INQUIRE_Q(args)
        except mq.MQMIError as e:
            if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_UNKNOWN_OBJECT_NAME:
                pass
            else:
                raise
        else:
            for queue_info in response:
                try:
                    q = queue_info[mq.CMQC.MQCA_Q_NAME]
                    args = {mq.CMQC.MQCA_Q_NAME: q,
                            mq.CMQC.MQIA_Q_TYPE: mq.CMQC.MQQT_LOCAL}
                    pcf.MQCMD_DELETE_Q(args)
                except mq.MQMIError:
                    pass
        try:
            sub_prefix = config.MQ.QUEUE.PREFIX + "*"
            args = {mq.CMQCFC.MQCACF_SUB_NAME: sub_prefix}
            response = pcf.MQCMD_INQUIRE_SUBSCRIPTION(args)
        except mq.MQMIError as e:
            if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_NO_SUBSCRIPTION:
                pass
            else:
                raise
        else:
            for sub_info in response:
                try:
                    sub = sub_info[mq.CMQCFC.MQBACF_SUB_ID]
                    args = {mq.CMQCFC.MQBACF_SUB_ID: mq.ByteString(sub)}
                    pcf.MQCMD_DELETE_SUBSCRIPTION(args)
                except mq.MQMIError:
                    pass

    def setUp(self):
        self.topic_string_template = "/UNITTEST/{prefix}/PUBSUB/{{type}}/{{destination}}/{{durable}}".format(
            prefix=config.MQ.QUEUE.PREFIX)
        self.subname_template = "{prefix}'s {{type}} {{destination}} {{durable}} Subscription".format(
            prefix=config.MQ.QUEUE.PREFIX)
        self.msg_template = "Hello World in the topic string \"{{topic_string}}\""
        # max length of queue names is 48 characters
        self.queue_name_template = "{prefix}_Q_TEST_PUBSUB_{{type}}_PROVIDED_{{durable}}".format(prefix=config.MQ.QUEUE.PREFIX)
        self.queue_manager = config.MQ.QM.NAME
        self.channel = config.MQ.QM.CHANNEL
        self.host = config.MQ.QM.HOST
        self.port = config.MQ.QM.PORT
        self.user = config.MQ.QM.USER
        self.password = config.MQ.QM.PASSWORD

        self.conn_info = "{0}({1})".format(self.host, self.port)

        self.qmgr = mq.QueueManager(None)
        self.qmgr.connectTCPClient(self.queue_manager, mq.CD(), self.channel, self.conn_info, self.user, self.password)

        # list of tuples (subscription, subscription descriptions) for tearDown() to delete after the test
        self.sub_desc_list = []

        # Some of the tests appear to leave dangling resources, especially after any failures. So we clean up
        # before each run
        self.cleanup()

    def msg_format(self, **kwargs):
        res = self.msg_template.format(**kwargs)
        return utils.py3str2bytes(res)

    def delete_sub(self, sub_desc):
        # can only delete a durable subscription
        if sub_desc["Options"] & mq.CMQC.MQSO_DURABLE:
            subname = sub_desc.get_vs("SubName")
            pcf = mq.PCFExecute(self.qmgr)
            args = {mq.CMQCFC.MQCACF_SUB_NAME: subname}
            pcf.MQCMD_DELETE_SUBSCRIPTION(args)

    def delete_queue(self, sub_desc, queue_name):
        # must be unmanaged
        if not sub_desc["Options"] & mq.CMQC.MQSO_MANAGED:
            pcf = mq.PCFExecute(self.qmgr)
            args = {mq.CMQC.MQCA_Q_NAME: utils.py3str2bytes(queue_name),
                    mq.CMQCFC.MQIACF_PURGE: mq.CMQCFC.MQPO_YES}
            pcf.MQCMD_DELETE_Q(args)

    def tearDown(self):
        """Delete the created objects.
        """
        for (sub, sub_desc, queue_name) in self.sub_desc_list:
            # self.delete_sub(sub_desc)
            if queue_name is None:
                sub_queue = sub.get_sub_queue()
                self.delete_queue(sub_desc, sub_queue)
            else:
                self.delete_queue(sub_desc, queue_name)
        self.qmgr.disconnect()

    @staticmethod
    def get_subscription_descriptor(subname, topic_string, options=0):
        sub_desc = mq.SD()
        sub_desc["Options"] = options
        sub_desc.set_vs("SubName", subname)
        sub_desc.set_vs("ObjectString", topic_string)
        return sub_desc

    def pub(self, msg, topic_string, *opts):
        topic = mq.Topic(self.qmgr, topic_string=topic_string)
        topic.open(open_opts=mq.CMQC.MQOO_OUTPUT)
        if not isinstance(msg, (str, bytes)):
            raise AttributeError('msg must be bytes or str to publish to topic.')  # py3
        topic.pub(msg, *opts)
        topic.close()

    def pub_rfh2(self, msg, topic_string, *opts):
        topic = mq.Topic(self.qmgr, topic_string=topic_string)
        topic.open(open_opts=mq.CMQC.MQOO_OUTPUT)
        if not isinstance(msg, (str, bytes)):
            raise AttributeError('msg must be bytes or str to publish to topic.')  # py3
        topic.pub_rfh2(msg, *opts)
        topic.close()

    def create_api_subscription(self):
        return mq.Subscription(self.qmgr)

    def create_admin_subscription(self, destination_class, subname, queue_name, topic_string):
        pcf = mq.PCFExecute(self.qmgr)
        args = {mq.CMQCFC.MQCACF_SUB_NAME: utils.py3str2bytes(subname),
                mq.CMQC.MQCA_TOPIC_STRING: utils.py3str2bytes(topic_string),
                mq.CMQCFC.MQIACF_DESTINATION_CLASS: destination_class,
                mq.CMQCFC.MQIACF_REPLACE: mq.CMQCFC.MQRP_YES}
        if destination_class is mq.CMQC.MQDC_PROVIDED:
            args[mq.CMQCFC.MQCACF_DESTINATION] = utils.py3str2bytes(queue_name)
        pcf.MQCMD_CREATE_SUBSCRIPTION(args)

    @staticmethod
    def create_get_opts():
        get_opts = mq.GMO(
            Options=mq.CMQC.MQGMO_NO_SYNCPOINT | mq.CMQC.MQGMO_FAIL_IF_QUIESCING | mq.CMQC.MQGMO_WAIT)
        get_opts["WaitInterval"] = 15000
        return get_opts

    def create_queue(self, queue_name):
        queue_type = mq.CMQC.MQQT_LOCAL
        max_depth = 123456

        args = {mq.CMQC.MQCA_Q_NAME: utils.py3str2bytes(queue_name),
                mq.CMQC.MQIA_Q_TYPE: queue_type,
                mq.CMQC.MQIA_MAX_Q_DEPTH: max_depth}
        pcf = mq.PCFExecute(self.qmgr)
        pcf.MQCMD_CREATE_Q(args)

############################################################################
#
# Real Tests start here
#
############################################################################

    def test_pubsub_api_managed_durable(self):
        topic_string = self.topic_string_template.format(type="API", destination="MANAGED", durable="DURABLE")
        subname = self.subname_template.format(type="Api", destination="Managed", durable="Durable")
        msg = self.msg_format(topic_string=topic_string)
        # register Subscription
        sub = self.create_api_subscription()

        # define a list self.sub_desc_list of subscription definitions so tearDown() can find it
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_DURABLE +
                                                    mq.CMQC.MQSO_MANAGED)
        self.sub_desc_list = [(sub, sub_desc, None)]

        sub.sub(sub_desc=sub_desc)
        # publish (put)
        self.pub(msg, topic_string)
        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)
        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_api_managed_durable_1_to_n(self):
        """Test multiple subscriptions."""
        # number of subscriptions
        nsub = 5
        topic_string = self.topic_string_template.format(type="API", destination="MANAGED", durable="DURABLE")
        msg = self.msg_format(topic_string=topic_string)
        self.sub_desc_list = []
        subscriptions = []
        for n in range(nsub):
            sub_desc = self.get_subscription_descriptor(
                self.subname_template.format(type="Api", destination="Managed", durable="Durable{0}".format(n)),
                self.topic_string_template.format(type="API", destination="MANAGED", durable="DURABLE"),
                mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_DURABLE + mq.CMQC.MQSO_MANAGED)
            # register Subscription
            sub = self.create_api_subscription()
            self.sub_desc_list.append((sub, sub_desc, None))
            sub.sub(sub_desc=sub_desc)
            subscriptions.append(sub)

        # publish (put)
        self.pub(msg, topic_string)

        get_opts = self.create_get_opts()
        for n in range(nsub):
            data = subscriptions[n].get(None, mq.md(), get_opts)
            subscriptions[n].close(sub_close_options=0, close_sub_queue=True)
            self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_api_managed_non_durable(self):
        topic_string = self.topic_string_template.format(type="API", destination="MANAGED", durable="NON DURABLE")
        subname = self.subname_template.format(type="Api", destination="Managed", durable="Non Durable")
        msg = self.msg_format(topic_string=topic_string)
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_MANAGED)
        # register Subscription
        sub = self.create_api_subscription()
        self.sub_desc_list = [(sub, sub_desc, None)]
        sub.sub(sub_desc=sub_desc)
        # publish (put)
        self.pub(msg, topic_string)
        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)
        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_api_managed_non_durable_rfh2(self):
        topic_string = self.topic_string_template.format(type="API_RFH2", destination="MANAGED", durable="NON DURABLE")
        subname = self.subname_template.format(type="Api_rfh2", destination="Managed", durable="Non Durable")
        msg = self.msg_format(topic_string=topic_string)
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_MANAGED)
        # register Subscription
        sub = self.create_api_subscription()
        self.sub_desc_list = [(sub, sub_desc, None)]
        sub.sub(sub_desc=sub_desc)

        # publish (put)
        put_mqmd = mq.md(Format=mq.CMQC.MQFMT_RF_HEADER_2,
                         Encoding=273,
                         CodedCharSetId=1208)

        put_opts = mq.pmo()

        put_rfh2 = mq.RFH2(_StrucId=mq.CMQC.MQRFH_STRUC_ID,
                              Version=mq.CMQC.MQRFH_VERSION_2,
                              StrucLength=188,
                              Encoding=273,
                              CodedCharSetId=1208,
                              Format=mq.CMQC.MQFMT_STRING,
                              Flags=0,
                              NameValueCCSID=1208)
        # pylint: disable=line-too-long
        put_rfh2.add_folder(b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc>")
        put_rfh2.add_folder(b"<testFolder><testVar>testValue</testVar></testFolder>")
        put_rfh2.add_folder(b"<mcd><Msd>xmlnsc</Msd></mcd>")

        self.pub_rfh2(msg, topic_string, put_mqmd, put_opts, [put_rfh2])
        get_opts = mq.GMO(Version=mq.CMQC.MQGMO_VERSION_4,
                             WaitInterval=15000,
                             Options=mq.CMQC.MQGMO_NO_SYNCPOINT |
                                    mq.CMQC.MQGMO_FAIL_IF_QUIESCING |
                                    mq.CMQC.MQGMO_WAIT)
        get_rfh2_list = []
        data = sub.get_rfh2(None, mq.md(Version=mq.CMQC.MQMD_VERSION_2), get_opts, get_rfh2_list)
        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_admin_managed(self):
        topic_string = self.topic_string_template.format(type="ADMIN", destination="MANAGED", durable="DURABLE")
        subname = self.subname_template.format(type="Admin", destination="Managed", durable="Durable")
        msg = self.msg_format(topic_string=topic_string)
        queue_name = self.queue_name_template.format(type="ADMIN", durable="DURABLE")
        sub_desc = self.get_subscription_descriptor(subname, topic_string, mq.CMQC.MQSO_RESUME)

        # register Subscription
        self.create_admin_subscription(mq.CMQC.MQDC_MANAGED, subname, queue_name, topic_string)
        sub = mq.Subscription(self.qmgr)
        self.sub_desc_list = [(sub, sub_desc, None)]
        sub.sub(sub_desc=sub_desc)
        # publish (put)
        self.pub(msg, topic_string)

        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)

        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_api_provided_durable(self):
        topic_string = self.topic_string_template.format(type="API", destination="PROVIDED", durable="DURABLE")
        subname = self.subname_template.format(type="Api", destination="Provided", durable="Durable")
        msg = self.msg_format(topic_string=topic_string)
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_DURABLE)
        queue_name = self.queue_name_template.format(type="API", durable="DURABLE")
        self.create_queue(queue_name)

        # create queue
        open_opts = mq.CMQC.MQOO_INPUT_AS_Q_DEF
        sub_queue = mq.Queue(self.qmgr, queue_name, open_opts)
        # register Subscription
        sub = self.create_api_subscription()
        self.sub_desc_list = [(sub, sub_desc, queue_name)]
        sub.sub(sub_desc=sub_desc, sub_queue=sub_queue)
        # publish (put)
        self.pub(msg, topic_string)

        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)

        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_api_provided_non_durable(self):
        topic_string = self.topic_string_template.format(type="API", destination="PROVIDED", durable="NON DURABLE")
        subname = self.subname_template.format(type="Api", destination="Provided", durable="None Durable")
        msg = self.msg_format(topic_string=topic_string)
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE)
        queue_name = self.queue_name_template.format(type="API", durable="NON_DURABLE")
        # create queue
        self.create_queue(queue_name)
        open_opts = mq.CMQC.MQOO_INPUT_AS_Q_DEF
        sub_queue = mq.Queue(self.qmgr, queue_name, open_opts)
        # register Subscription
        sub = self.create_api_subscription()
        sub.sub(sub_desc=sub_desc, sub_queue=sub_queue)
        self.sub_desc_list = [(sub, sub_desc, queue_name)]
        # publish (put)
        self.pub(msg, topic_string)
        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)
        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_admin_provided(self):
        topic_string = self.topic_string_template.format(type="ADMIN", destination="PROVIDED", durable="DURABLE")
        subname = self.subname_template.format(type="Admin", destination="Provided", durable="Durable")
        msg = self.msg_format(topic_string=topic_string)
        queue_name = self.queue_name_template.format(type="ADMIN", durable="DURABLE")
        sub_desc = self.get_subscription_descriptor(subname, topic_string, mq.CMQC.MQSO_RESUME)
        # create queue
        self.create_queue(queue_name)
        open_opts = mq.CMQC.MQOO_INPUT_AS_Q_DEF
        sub_queue = mq.Queue(self.qmgr, queue_name, open_opts)

        # register Subscription
        self.create_admin_subscription(mq.CMQC.MQDC_PROVIDED, subname, queue_name, topic_string)
        sub = mq.Subscription(self.qmgr)

        sub.sub(sub_desc=sub_desc, sub_queue=sub_queue)
        self.sub_desc_list = [(sub, sub_desc, queue_name)]
        # publish (put)
        self.pub(msg, topic_string)

        get_opts = self.create_get_opts()
        data = sub.get(None, mq.md(), get_opts)

        sub.close(sub_close_options=0, close_sub_queue=True)
        self.assertTrue(utils.strcmp(data, msg))

    def test_pubsub_already_exists(self):
        """Trying to register an already existing subscription should raise an exception.
        """
        topic_string = self.topic_string_template.format(type="API", destination="MANAGED", durable="DURABLE")
        subname = self.subname_template.format(type="Api", destination="Managed", durable="Durable")

        # define a list self.sub_desc_list of subscription definitions so tearDown() can find it
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_DURABLE + mq.CMQC.MQSO_MANAGED)

        # register Subscription
        sub = self.create_api_subscription()
        # this modifies the subscription descriptor
        sub.sub(sub_desc=sub_desc)
        sub = self.create_api_subscription()
        self.sub_desc_list = [(sub, sub_desc, None)]
        with self.assertRaises(mq.MQMIError) as cm:
            # create a new subscription descriptor
            # but do not add it to the list self.sub_desc_list
            # because tearDown() would try to delete the subscription
            # and fail because this registration will not succeed
            sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                        mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_DURABLE +
                                                        mq.CMQC.MQSO_MANAGED)
            sub.sub(sub_desc=sub_desc)
        # Exception should be
        # FAILED: MQRC_SUB_ALREADY_EXISTS
        self.assertEqual(cm.exception.reason, mq.CMQC.MQRC_SUB_ALREADY_EXISTS)

    def test_pubsub_encoding(self):
        """Test Encoding in managed and non durable subscription.
        """
        topic_string = self.topic_string_template.format(type="API", destination="MANAGED", durable="NON DURABLE")
        subname = self.subname_template.format(type="Api", destination="Managed", durable="Non Durable")
        messages = ["ascii", unicode("Euro sign: �", "iso-8859-15"), unicode("Uml�ut", "iso-8859-15"), unicodedata.lookup("INFINITY")]

        md = mq.md()
        # setting this means the message is entirely character data
        # md.Format = mq.CMQC.MQFMT_STRING
        # default
        # md.CodedCharSetId = mq.CMQC.MQCCSI_Q_MGR
        # UTF-8
        md.CodedCharSetId = 1208
        # UCS-2
        # md.CodedCharSetId = 1200
        # ISO-8859-1
        # md.CodedCharSetId = 819
        # ASCII
        # md.CodedCharSetId = 437

        # do not add the subscription to the list,
        # because tearDown() does not have to delete the subscription (in this non durable case)
        sub_desc = self.get_subscription_descriptor(subname, topic_string,
                                                    mq.CMQC.MQSO_CREATE + mq.CMQC.MQSO_MANAGED)
        # register Subscription
        sub = self.create_api_subscription()
        sub.sub(sub_desc=sub_desc)
        # publish (put)
        for msg in messages:
            self.pub(msg.encode("utf-8"), topic_string, md)

        get_opts = self.create_get_opts()
        get_opts["Options"] += mq.CMQC.MQGMO_CONVERT
        # md.CodedCharSetId = 819
        # md.CodedCharSetId = 437
        for msg in messages:
            # clear md for re-use
            md.MsgId = mq.CMQC.MQMI_NONE
            md.CorrelId = mq.CMQC.MQCI_NONE
            md.GroupId = mq.CMQC.MQGI_NONE
            # md.CodedCharSetId = 819
            data = sub.get(None, md, get_opts)
            self.assertEqual(unicode(data, "utf-8"), msg)
        sub.close(sub_close_options=0, close_sub_queue=True)


if __name__ == "__main__":
    unittest.main()



"""MQXQH: Transmission header"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, MD

class XQH(MQOpts):
    """ Construct an MQXQH Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    Note that this is different from the C MQI definition of the XQH, as that incorporates
    an embedded MQMDv1 structure. Instead, we have the get_embedded_md() method to extract
    that separately. This is primarily used to parse an already-available XQH, not to build a new
    message from scratch.

    Note that if an app really wants to construct an XQH to PUT a message directly to an XMITQ,
    then you'd have to do quite a bit more work which we don't really assist with in this class.
    But normally, the XQH will come automatically when you write to a remote queue.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQXQH_STRUC_ID, '4s'],
                ['Version', CMQC.MQXQH_VERSION_1, MQLONG_TYPE],
                ['RemoteQName', b'', '48s'],
                ['RemoteQMgrName', b'', '48s'], ]

        super().__init__(tuple(opts), **kw)

    # Return the XQH as a formatted structure based on the message buffer.
    def get_header(self, buf):
        """Return the unpacked header from the full message buffer"""
        xqh = super().unpack(buf[:CMQC.MQXQH_CURRENT_LENGTH - CMQC.MQMD_LENGTH_1])
        return xqh

    # Return the MQMD that is part of the real (full) MQXQH structure. It is always
    # an MQMDv1, but this package usually deals with MQMDv2. So we extract the block and add
    # a zero-filled pad array to ensure it's the right length. Which means that the MDv2 fields
    # (eg GroupId) may not be accurate if the original message was using those MDv2 fields. So we
    # remove them from the returned MD.
    # If the Format in this MD indicates there is a subsequent MQMDE, which corresponds to
    # the MDv2 fields, then you can extract that and unpack it directly before then
    # continuing on to the rest of the message body.
    def get_embedded_md(self, buf):
        """Return the unpacked MQMDv1 header from the full message buffer"""
        offset = CMQC.MQXQH_CURRENT_LENGTH - CMQC.MQMD_LENGTH_1         # Where does it start in the buffer
        pad = CMQC.MQMD_LENGTH_2 - CMQC.MQMD_LENGTH_1                   # How long is the padding
        md_buf = buf[offset:CMQC.MQXQH_CURRENT_LENGTH] + bytearray(pad)
        md = MD().unpack(md_buf)
        # Remove the MDv2-specific fields.
        md._remove('GroupId')
        md._remove('MsgSeqNumber')
        md._remove('Offset')
        md._remove('MsgFlags')
        md._remove('OriginalLength')
        return md



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how to build a PCF command to set a user's authority.
The "principal" is set and will always work on systems where UserExternal is configured, allowing
unknown userids to be used in these commands. Otherwise, additional configuration would
be required on the queue manager to define users or groups; something we want to avoid in
examples wherever possible.
"""

import logging

import ibmmq as mq

logging.basicConfig(level=logging.INFO)

queue_manager = 'QM1'
channel = 'DEV.ADMIN.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'admin'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)
pcf = mq.PCFExecute(qmgr)

principal_entity = [b'app']
authorities = [mq.CMQCFC.MQAUTH_BROWSE,
               mq.CMQCFC.MQAUTH_INQUIRE]

args = {mq.CMQCFC.MQCACF_AUTH_PROFILE_NAME: 'DEV.QUEUE.1',
        mq.CMQCFC.MQIACF_OBJECT_TYPE: mq.CMQC.MQOT_Q,
        mq.CMQCFC.MQIACF_AUTH_ADD_AUTHS: authorities,
        mq.CMQCFC.MQCACF_PRINCIPAL_ENTITY_NAMES: principal_entity}

result = pcf.MQCMD_SET_AUTH_REC(args)

qmgr.disconnect()

print("Done.")



"""MQSRO: Subscription Request Options"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class SRO(MQOpts):
    """ Construct an MQSRO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQSRO_STRUC_ID, '4s'],
                ['Version', CMQC.MQSRO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQSRO_FAIL_IF_QUIESCING, MQLONG_TYPE],
                ['NumPubs', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example sets up a TLS connection to the queue manager. It requires that
a suitable channel has been defined, and that certificates are in the appropriate
keystores, locally and on the qmgr.
"""
import logging

import ibmmq as mq

logging.basicConfig(level=logging.INFO)

queue_manager = 'QM1'
channel = 'SYSTEM.SSL.SVRCONN.1'
host = '127.0.0.1'
port = '1414'

conn_info = '%s(%s)' % (host, port)
ssl_cipher_spec = 'ANY_TLS12_OR_HIGHER'
key_repo_location = '/var/mqm/ssl/key'
message = 'Hello from Python!'

queue_name = 'DEV.QUEUE.1'

cd = mq.CD()
cd.ChannelName = channel
cd.ConnectionName = conn_info
cd.ChannelType = mq.CMQXC.MQCHT_CLNTCONN
cd.TransportType = mq.CMQXC.MQXPT_TCP
cd.SSLCipherSpec = ssl_cipher_spec

sco = mq.SCO()
sco.KeyRepository = key_repo_location

# Force a client connection
cno = mq.CNO()
cno.Options = mq.CMQC.MQCNO_CLIENT_BINDING

qmgr = mq.QueueManager(None)
qmgr.connect_with_options(queue_manager, cd, sco, cno=cno)
print("Connection successful.")

put_queue = mq.Queue(qmgr, queue_name)
put_queue.put(message)

get_queue = mq.Queue(qmgr, queue_name)
logging.info('Here is the message again: [%s]', get_queue.get())

put_queue.close()
get_queue.close()
qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows reading a single message from a queue, waiting
for up to 5 seconds before giving up.
"""

import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'
WAIT_INTERVAL = 5  # seconds

# Message Descriptor
md = mq.MD()

# Get Message Options
gmo = mq.GMO()
gmo.Options = mq.CMQC.MQGMO_WAIT | mq.CMQC.MQGMO_FAIL_IF_QUIESCING
gmo.WaitInterval = WAIT_INTERVAL * 1000

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

queue = mq.Queue(qmgr, queue_name)
print("Waiting for message ...")
message = queue.get(None, md, gmo)
print("Message: ", message)
queue.close()

qmgr.disconnect()



"""Queue class: implements Queue-based MQI verbs including
MQOPEN, MQPUT, MQGET, MQCLOSE, MQCB, MQINQ, MQSET
"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.
from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, PMO, GMO, RFH2, ibmmqc, MQObject
from mqqmgr import *

import mqinq
import mqqargs

import mqcallback

import mqlog

# unicode = str

class Queue(MQObject):
    """ Queue encapsulates all the Queue I/O operations, including
    open/close and get/put. A QueueManager object must be already
    connected. The Queue may be opened implicitly on construction, or
    the open may be deferred until a call to open(), put() or
    get(). The Queue to open is identified either by a queue name
    string (in which case a default MQOD structure is created using
    that name), or by passing a ready constructed MQOD class.
    """
    def __real_open(self):
        """Really open the queue."""
        mqlog.trace_entry("queue:__real_open")
        if self.__q_desc is None:
            mqlog.trace_exit("queue:__real_open", ep=1)
            raise PYIFError('The Queue Descriptor has not been set.')
        rv = ibmmqc.MQOPEN(self.__q_mgr.get_handle(), self.__q_desc.pack(), self.__open_opts)
        if rv[-2]:
            mqlog.trace_exit("queue:__real_open", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        self.__q_handle = rv[0]
        self.__q_desc.unpack(rv[1])

        if OTelFunctions.open:
            OTelFunctions.open(self, self.__q_desc, self.__open_opts, None)
        mqlog.trace_exit("queue:__real_open")

    def __init__(self, qmgr: QueueManager, *opts):
        """ Associate a Queue instance with the QueueManager object 'qmgr'
        and optionally open the Queue.

        If q_desc is passed, it identifies the Queue either by name (if
        its a string), or by MQOD (if its a OD() instance). If
        q_desc is not defined, then the Queue is not opened
        immediately, but deferred to a subsequent call to open().

        If openOpts is passed, it specifies queue open options, and
        the queue is opened immediately. If open_opts is not passed,
        the queue open is deferred to a subsequent call to open(),
        put() or get().

        The following table clarifies when the Queue is opened:

           qDesc  openOpts   When opened
             N       N       open()
             Y       N       open() or get() or put()
             Y       Y       Immediately
        """
        mqlog.trace_entry("queue:__init__")

        self.__q_mgr = qmgr
        self.__q_handle = self.__q_desc = self.__open_opts = None
        ln = len(opts)
        if ln > 2:
            mqlog.trace_exit("queue:__init__", ep=1)
            raise TypeError('Too many args')
        if ln > 0:
            self.__q_desc = mqqargs._make_q_desc(opts[0])
        if ln == 2:
            self.__open_opts = opts[1]
            self.__real_open()

        if self.__q_desc:
            try:
                q_name = self.__q_desc.ObjectName
            except AttributeError:
                q_name = ""
        else:
            q_name = ""
        super().__init__(q_name)
        mqlog.trace_exit("queue:__init__")

    def __del__(self):
        """ Close the queue, if it has been opened.
        """
        mqlog.trace_entry("queue:__del__")
        if self.__q_handle:
            try:
                self.close()
            except (PYIFError, MQMIError):
                pass
        mqlog.trace_exit("queue:__del__")

    def open(self, q_desc, *opts):
        """ Open the queue specified by q_desc. q_desc identifies the Queue
        either by name (if its a string), or by MQOD (if its a
        OD() instance). If openOpts is passed, it defines the
        queue open options, and the Queue is opened immediately. If
        openOpts is not passed, the Queue open is deferred until a
        subsequent put() or get() call.
        """
        mqlog.trace_entry("queue:open")

        ln = len(opts)
        if ln > 1:
            mqlog.trace_exit("queue:open", ep=1)
            raise TypeError('Too many args')
        if self.__q_handle:
            mqlog.trace_exit("queue:open", ep=2)
            raise PYIFError('The Queue is already open')
        self.__q_desc = mqqargs._make_q_desc(q_desc)
        if ln == 1:
            self.__open_opts = opts[0]
            self.__real_open()
        mqlog.trace_exit("queue:open")

    def put(self, msg, *opts):
        """ Put the string buffer 'msg' on the queue. If the queue is not
        already open, it is opened now with the option 'MQOO_OUTPUT'.

        m_desc is the MQ() Message Descriptor for the
        message. If it is not passed, or is None, then a default md()
        object is used.

        put_opts is the PMO() Put Message Options structure
        for the put call. If it is not passed, or is None, then a
        default pmo() object is used.

        If m_desc and/or put_opts arguments were supplied, they may be
        updated by the put operation.
        """

        mqlog.trace_entry("queue:put")

        m_desc, put_opts = mqqargs.common_q_args(*opts)
        if not isinstance(msg, bytes):
            if isinstance(msg, str):
                msg = msg.encode(self.__q_mgr.bytes_encoding)
                m_desc.CodedCharSetId = self.__q_mgr.default_ccsid
                m_desc.Format = CMQC.MQFMT_STRING
            else:
                error_message = 'Message type is {0}. Convert to bytes.'
                mqlog.trace_exit("queue:put", ep=1)
                raise TypeError(error_message.format(type(msg)))

        if put_opts is None:
            put_opts = PMO()

        # If queue open was deferred, open it for put now
        if not self.__q_handle:
            self.__open_opts = CMQC.MQOO_OUTPUT
            self.__real_open()

        if OTelFunctions.put_trace_before:
            OTelFunctions.put_trace_before(self.__q_mgr, m_desc, put_opts, msg)

        # Now send the message
        rv = ibmmqc.MQPUT(self.__q_mgr.get_handle(), self.__q_handle, m_desc.pack(), put_opts.pack(), msg)
        if rv[-2]:
            mqlog.trace_exit("queue:put", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])
        _ = m_desc.unpack(rv[0])
        _ = put_opts.unpack(rv[1])

        if OTelFunctions.put_trace_after:
            OTelFunctions.put_trace_after(self.__q_mgr, put_opts)
        mqlog.trace_exit("queue:put")

    def put_rfh2(self, msg, *opts):
        """ Put a RFH2 message. opts[2] is a list of RFH2 headers. MQMD and RFH2's must be correct.
        """
        mqlog.trace_entry("queue:put_rfh2")

        ensure_not_unicode(msg)  # Python 3 bytes check

        rfh2_buff = b''
        if len(opts) >= 3:
            if opts[2] is not None:
                if not isinstance(opts[2], list):
                    mqlog.trace_exit("queue:put_rfh2", ep=1)
                    raise TypeError('Third item of opts should be a list.')
                encoding = CMQC.MQENC_NATIVE
                if opts[0] is not None:
                    mqmd = opts[0]
                    encoding = mqmd['Encoding']

                for rfh2_header in opts[2]:
                    if rfh2_header is not None:
                        rfh2_buff = rfh2_buff + rfh2_header.pack(encoding)
                        encoding = rfh2_header['Encoding']

                msg = rfh2_buff + msg
            self.put(msg, *opts[0:2])
        else:
            self.put(msg, *opts)
        mqlog.trace_exit("queue:put_rfh2")

    # pylint complains about a couple of things here which we will ignore in order to maintain
    # backwards compatibility
    def get(self, maxLength=None, *opts, **kwargs):  # pylint: disable=invalid-name,keyword-arg-before-vararg
        """ Return a message from the queue. If the queue is not already
        open, it is opened now with the option 'MQOO_INPUT_AS_Q_DEF'.

        maxLength, if present, specifies the maximum length for the
        message. If the message received exceeds maxLength, then the
        behavior is as defined by MQI and the get_opts argument.

        If maxLength is not specified, or is None, then the entire
        message is returned regardless of its size. This may require
        multiple calls to the underlying MQGET API. Other get() variants
        use max_length as the parameter, but this one has to stay in camelCase
        for backwards compatibility.

        m_desc is the MD() Message Descriptor for receiving
        the message. If it is not passed, or is None, then a default
        MD() object is used.

        get_opts is the GMO() Get Message Options
        structure for the get call. If it is not passed, or is None,
        then a default GMO() object is used.

        If m_desc and/or get_opts arguments were supplied, they may be
        updated by the get operation.

        If you want to supply the MD or GMO values, then you must also
        supply a maxLength value.
        """
        mqlog.trace_entry("queue:get")

        removed = 0

        otel_options = kwargs['otel_options'] if 'otel_options' in kwargs else None
        if otel_options and not isinstance(otel_options, OTelOptions):
            mqlog.trace_exit("queue:get", ep=1)
            raise TypeError('otel_options must be instance of OTelOptions.')

        max_length = maxLength  # Work with the "right" name style from here on
        m_desc, get_opts = mqqargs.common_q_args(*opts)
        if get_opts is None:
            get_opts = GMO()

        # If queue open was deferred, open it for put now
        if not self.__q_handle:
            self.__open_opts = CMQC.MQOO_INPUT_AS_Q_DEF
            self.__real_open()

        if max_length is None:
            if get_opts.Options & CMQC.MQGMO_ACCEPT_TRUNCATED_MSG:
                length = 0
            else:
                length = 4096  # Try to read short message in one call
        else:
            length = max_length

        if OTelFunctions.get_trace_before:
            OTelFunctions.get_trace_before(self.__q_mgr, self, get_opts, False)

        rv = ibmmqc.MQGET(self.__q_mgr.get_handle(), self.__q_handle, m_desc.pack(), get_opts.pack(), length)

        if not rv[-2]:
            # Everything is OK
            _ = m_desc.unpack(rv[1])
            _ = get_opts.unpack(rv[2])

            if OTelFunctions.get_trace_after:
                removed = OTelFunctions.get_trace_after(self, get_opts, m_desc, otel_options, rv[0], False)
            # This is the most common route, so no "ep" on this trace point
            mqlog.trace_exit("queue:get")
            return rv[0][removed:]

        # Accept truncated message
        if ((rv[-1] == CMQC.MQRC_TRUNCATED_MSG_ACCEPTED) or
                # Do not reread message with original length
                (rv[-1] == CMQC.MQRC_TRUNCATED_MSG_FAILED and max_length is not None) or
                # Other errors
                (rv[-1] != CMQC.MQRC_TRUNCATED_MSG_FAILED)):
            if rv[-2] == CMQC.MQCC_WARNING:
                _ = m_desc.unpack(rv[1])
                _ = get_opts.unpack(rv[2])
            mqlog.trace_exit("queue:get", ep=3, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1], message=rv[0], original_length=rv[-3])

        # Message truncated, but we know its size. Do another MQGET
        # to retrieve it from the queue.
        rv = ibmmqc.MQGET(self.__q_mgr.get_handle(), self.__q_handle, m_desc.pack(), get_opts.pack(), rv[-3])
        if rv[-2]:
            mqlog.trace_exit("queue:get", ep=4, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        _ = m_desc.unpack(rv[1])
        _ = get_opts.unpack(rv[2])

        # Only process OTel tracing if we actually got a message
        if rv[-2] == CMQC.MQCC_OK or rv[-1] == CMQC.MQRC_TRUNCATED_MSG_ACCEPTED:
            if OTelFunctions.get_trace_after:
                removed = OTelFunctions.get_trace_after(self, get_opts, m_desc, otel_options, rv[0], False)

        if rv[0]:
            mqlog.trace_exit("queue:get", ep=5)
            return rv[0][removed:]

        mqlog.trace_exit("queue:get", ep=6)
        return rv[0]

    def get_no_jms(self, max_length=None, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        """Get a message but force there to be no properties returned."""
        mqlog.trace_entry("queue:get_no_jms")

        md, gmo = mqqargs.common_q_args(*args)
        if not gmo:
            gmo = GMO()
        gmo.Options = gmo.Options | CMQC.MQGMO_NO_PROPERTIES | CMQC.MQGMO_FAIL_IF_QUIESCING

        rv = self.get(max_length, md, gmo, kwargs)
        mqlog.trace_exit("queue:get_no_jms")
        return rv

    get_no_rfh2 = get_no_jms

    def get_rfh2(self, max_length=None, *opts, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        """ Get a message and attempt to unpack the rfh2 headers.
        opts[2] should be a empty list.
        Unpacking only attempted if Format in previous header is
        CMQC.MQFMT_RF_HEADER_2.
        """
        mqlog.trace_entry("queue:get_rfh2")

        if len(opts) >= 3:
            if opts[2] is not None:
                if not isinstance(opts[2], list):
                    mqlog.trace_exit("queue:get_rfh2", ep=1)
                    raise TypeError('Third item of opts should be a list.')

                msg = self.get(max_length, *opts[0:2], kwargs)
                mqmd = opts[0]
                rfh2_headers = []
                # If format is not CMQC.MQFMT_RF_HEADER_2 then do not parse.
                frmt = mqmd['Format']
                while frmt == CMQC.MQFMT_RF_HEADER_2:
                    rfh2_header = RFH2()
                    rfh2_header.unpack(msg)
                    rfh2_headers.append(rfh2_header)
                    msg = msg[rfh2_header['StrucLength']:]
                    frmt = rfh2_header['Format']
                opts[2].extend(rfh2_headers)
            else:
                mqlog.trace_exit("queue:get_rfh2", ep=2)
                raise AttributeError('get_opts cannot be None if passed.')
        else:
            msg = self.get(max_length, *opts, kwargs)

        mqlog.trace_exit("queue:get_rfh2")

        return msg

    def close(self, options: int = CMQC.MQCO_NONE) -> None:
        """ Close a queue, using options.
        """
        mqlog.trace_entry("queue:close")

        if not self.__q_handle:
            mqlog.trace_exit("queue:close", ep=1)
            raise PYIFError('not open')
        rv = ibmmqc.MQCLOSE(self.__q_mgr.get_handle(), self.__q_handle, options)
        if rv[0]:
            mqlog.trace_exit("queue:close", ep=1, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        if OTelFunctions.close:
            OTelFunctions.close(self)

        self.__q_handle = self.__q_desc = self.__open_opts = None
        mqlog.trace_exit("queue:close")

    def inquire(self, selectors: Union[int, list[int]]) -> Union[Any, dict[int, Any]]:
        """ Inquire on queue attributes. If the queue is not already
        open, it is opened for Inquire.

        If the selectors parameter is a single value, then that specific
        attribute's value is returned (string or int).

        If the selectors parameter is a list of values, then a dict is returned
        where all the values are stored using each element of the selectors as the keys.
        """

        if not self.__q_handle:
            self.__open_opts = CMQC.MQOO_INQUIRE
            self.__real_open()
        # mqinq.inq will throw the exception if necessary
        rv = mqinq.common_inq(self.__q_mgr.get_handle(), self.__q_handle, selectors)
        return rv

    # Create an alias that is closer to the real MQI function name
    inq = inquire

    def set(self, *args) -> None:
        """ Sets a queue's attributes

        Old-style interface has two parameters, selector and value.
        Newer interface has a single parameter, a dict that maps between selectors and values.
        """
        mqlog.trace_entry("queue:set")

        if not self.__q_handle:
            self.__open_opts = CMQC.MQOO_SET
            self.__real_open()
        rv = mqinq.common_set(self.__q_mgr.get_handle(), self.__q_handle, args)

        if rv[1]:
            mqlog.trace_exit("queue:set", ep=1, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])
        mqlog.trace_exit("queue:set")

    def set_handle(self, queue_handle):
        """ Sets the queue handle in the case when a handle was returned from a previous MQ call.
        """
        self.__q_handle = queue_handle

    def get_handle(self) -> int:
        """ Get the queue handle.
        """
        return self.__q_handle

    def get_name(self) -> str:
        """Return the name of the queue as a stripped string"""
        v = self.__q_desc.ObjectName
        return self.to_string(v)

    def get_queue_manager(self) -> QueueManager:
        """ Get the queue manager object.
        """
        return self.__q_mgr

    # Setting up a callback - pass it to the function
    # common to both Queues and QMgrs
    def cb(self, **kwargs: dict[str, Any]) -> None:
        """cb(operation=operation, md=mqmd,gmo=mqgmo,cbd=mqcbd)
        Register or Deregister a Callback function for asynchronous
        message consumption.

        The cbd.CallbackFunction must be defined as (dict[str,Any]) with
        entries for queue_manager,queue,md,gmo,cbc,msg.
        """
        mqlog.trace_entry("queue:cb")
        mqcallback.real_cb(self, kwargs)
        mqlog.trace_exit("queue:cb")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows reading all the messages on a queue.
"""

import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'
WAIT_INTERVAL = 3  # seconds

# Message Descriptor
md = mq.MD()

# Get Message Options
gmo = mq.GMO()
gmo.Options = mq.CMQC.MQGMO_WAIT | mq.CMQC.MQGMO_FAIL_IF_QUIESCING
gmo.MatchOptions = mq.CMQC.MQMO_NONE
gmo.WaitInterval = WAIT_INTERVAL * 1000

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)
queue = mq.Queue(qmgr, queue_name)

keep_running = True

while keep_running:
    try:
        # Wait up to to gmo.WaitInterval for a new message.
        message = queue.get(None, md, gmo)

        # Process the message here..
        print("Message: ", message)

        # Reset the MsgId, CorrelId & GroupId so that we can reuse
        # the same 'md' object again.
        md.MsgId = mq.CMQC.MQMI_NONE
        md.CorrelId = mq.CMQC.MQCI_NONE
        md.GroupId = mq.CMQC.MQGI_NONE

    except mq.MQMIError as e:
        if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_NO_MSG_AVAILABLE:
            # No messages, that's OK, we can ignore it.
            print("No more messages.")
            keep_running = False
        else:
            # Some other error condition.
            raise

queue.close()
qmgr.disconnect()
